package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.MaterialTypeDto
import dev.fyloz.colorrecipesexplorer.dtos.MixTypeDto
import dev.fyloz.colorrecipesexplorer.service.MixTypeService
import org.springframework.transaction.annotation.Transactional

interface MixTypeLogic : Logic<MixTypeDto, MixTypeService> {
    /** Returns a mix type for the given [name] and [materialType]. If this mix type does not already exist, it will be created. */
    fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialTypeDto): MixTypeDto

    /** Updates the [mixType] with the given [name] and [materialType], or create a new one if it is shared with other mixes. */
    fun updateOrCreateForNameAndMaterialType(
        mixType: MixTypeDto,
        name: String,
        materialType: MaterialTypeDto
    ): MixTypeDto
}

@LogicComponent
class DefaultMixTypeLogic(service: MixTypeService) : BaseLogic<MixTypeDto, MixTypeService>(service, Constants.ModelNames.MIX_TYPE), MixTypeLogic {
    @Transactional
    override fun getOrCreateForNameAndMaterialType(name: String, materialType: MaterialTypeDto) =
        service.getByNameAndMaterialType(name, materialType.id) ?: saveForNameAndMaterialType(name, materialType)

    override fun updateOrCreateForNameAndMaterialType(
        mixType: MixTypeDto,
        name: String,
        materialType: MaterialTypeDto
    ) = if (service.existsByNameAndMaterialType(name, materialType.id, mixType.id)) {
        service.getByNameAndMaterialType(name, materialType.id)!!
    } else if (service.isShared(mixType.id)) {
        saveForNameAndMaterialType(name, materialType)
    } else {
        updateForNameAndMaterialType(mixType, name, materialType)
    }

    override fun deleteById(id: Long) {
        if (service.isUsedByMixes(id)) {
            throw cannotDeleteException("Cannot delete the mix type with the id '$id' because one or more mixes depends on it")
        }

        super.deleteById(id)
    }

    private fun saveForNameAndMaterialType(name: String, materialType: MaterialTypeDto): MixTypeDto {
        return save(MixTypeDto(name = name, materialType = materialType))
    }

    private fun updateForNameAndMaterialType(
        mixType: MixTypeDto,
        name: String,
        materialType: MaterialTypeDto
    ): MixTypeDto {
        return update(mixType.copy(name = name, materialType = materialType, material = mixType.material))
    }
}