package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.dtos.TouchUpKitDto
import dev.fyloz.colorrecipesexplorer.logic.config.ConfigurationLogic
import dev.fyloz.colorrecipesexplorer.logic.files.WriteableFileLogic
import dev.fyloz.colorrecipesexplorer.model.ConfigurationType
import dev.fyloz.colorrecipesexplorer.service.TouchUpKitService
import dev.fyloz.colorrecipesexplorer.utils.*
import org.springframework.core.io.ByteArrayResource
import org.springframework.core.io.Resource
import java.time.LocalDate

interface TouchUpKitLogic : Logic<TouchUpKitDto, TouchUpKitService> {
    /** Sets the touch up kit with the given [id] as complete. */
    fun complete(id: Long)

    /**
     * Generates and returns a [PdfDocument] for the given [job] as a [ByteArrayResource].
     *
     * If TOUCH_UP_KIT_CACHE_PDF is enabled and a file exists for the job, its content will be returned.
     * If caching is enabled but no file exists for the job, the generated ByteArrayResource will be cached on the disk.
     */
    fun generateJobPdfResource(job: String): Resource

    /** Generates and returns a [PdfDocument] for the given [job]. */
    fun generateJobPdf(job: String): PdfDocument

    /** Writes the given [pdf] to the disk if TOUCH_UP_KIT_CACHE_PDF is enabled. */
    fun cacheJobPdf(job: String, pdf: PdfDocument)
}

@LogicComponent
class DefaultTouchUpKitLogic(
    service: TouchUpKitService,
    private val fileLogic: WriteableFileLogic,
    private val configLogic: ConfigurationLogic
) : BaseLogic<TouchUpKitDto, TouchUpKitService>(service, Constants.ModelNames.TOUCH_UP_KIT), TouchUpKitLogic {
    private val cacheGeneratedFiles by lazy {
        configLogic.getContent(ConfigurationType.TOUCH_UP_KIT_CACHE_PDF) == true.toString()
    }

    override fun complete(id: Long) = service.updateCompletionDateById(id, LocalDate.now())

    override fun generateJobPdfResource(job: String): Resource {
        if (cacheGeneratedFiles) {
            val pdfPath = jobPdfPath(job)
            if (fileLogic.exists(pdfPath)) {
                return fileLogic.read(pdfPath)
            }
        }

        val pdf = generateJobPdf(job)
        cacheJobPdf(job, pdf)

        return pdf.toByteArrayResource()
    }

    override fun generateJobPdf(job: String) = pdf {
        container {
            centeredVertically = true
            drawContainerBottom = true
            text(TOUCH_UP_TEXT_FR) {
                bold = true
                fontSize = PDF_DEFAULT_FONT_SIZE + 12
            }
            text(TOUCH_UP_TEXT_EN) {
                bold = true
                fontSize = PDF_DEFAULT_FONT_SIZE + 12
            }
            text(job) {
                marginTop = 10f
            }
        }

        container(containers[0]) {
            drawContainerBottom = false
        }
    }

    override fun cacheJobPdf(job: String, pdf: PdfDocument) {
        if (!cacheGeneratedFiles) return

        fileLogic.write(pdf.toByteArrayResource(), jobPdfPath(job), true)
    }

    private fun jobPdfPath(job: String) =
        "${Constants.FilePaths.TOUCH_UP_KITS}/$job.pdf"

    companion object {
        const val TOUCH_UP_TEXT_FR = "KIT DE RETOUCHE"
        const val TOUCH_UP_TEXT_EN = "TOUCH UP KIT"
    }
}
