package dev.fyloz.colorrecipesexplorer.logic.users

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.LogicComponent
import dev.fyloz.colorrecipesexplorer.config.security.defaultGroupCookieName
import dev.fyloz.colorrecipesexplorer.dtos.GroupDto
import dev.fyloz.colorrecipesexplorer.dtos.UserDto
import dev.fyloz.colorrecipesexplorer.exception.NoDefaultGroupException
import dev.fyloz.colorrecipesexplorer.logic.BaseLogic
import dev.fyloz.colorrecipesexplorer.logic.Logic
import dev.fyloz.colorrecipesexplorer.service.GroupService
import org.springframework.transaction.annotation.Transactional
import org.springframework.web.util.WebUtils
import javax.servlet.http.HttpServletRequest
import javax.servlet.http.HttpServletResponse

const val defaultGroupCookieMaxAge = 10 * 365 * 24 * 60 * 60 // 10 ans

interface GroupLogic : Logic<GroupDto, GroupService> {
    /** Gets all the users of the group with the given [id]. */
    fun getUsersForGroup(id: Long): Collection<UserDto>

    /** Gets the default group from a cookie in the given HTTP [request]. */
    fun getRequestDefaultGroup(request: HttpServletRequest): GroupDto

    /** Sets the default group cookie for the given HTTP [response]. */
    fun setResponseDefaultGroup(id: Long, response: HttpServletResponse)
}

@LogicComponent
class DefaultGroupLogic(service: GroupService, private val userLogic: UserLogic) :
    BaseLogic<GroupDto, GroupService>(service, Constants.ModelNames.GROUP),
    GroupLogic {
    override fun getUsersForGroup(id: Long) = userLogic.getAllByGroup(getById(id))

    override fun getRequestDefaultGroup(request: HttpServletRequest): GroupDto {
        val defaultGroupCookie = WebUtils.getCookie(request, defaultGroupCookieName)
            ?: throw NoDefaultGroupException()
        val defaultGroupUser = userLogic.getById(
            defaultGroupCookie.value.toLong(),
            isSystemUser = false,
            isDefaultGroupUser = true
        )
        return defaultGroupUser.group!!
    }

    override fun setResponseDefaultGroup(id: Long, response: HttpServletResponse) {
        val defaultGroupUser = userLogic.getDefaultGroupUser(getById(id))
        response.addHeader(
            "Set-Cookie",
            "$defaultGroupCookieName=${defaultGroupUser.id}; Max-Age=$defaultGroupCookieMaxAge; Path=/api; HttpOnly; Secure; SameSite=strict"
        )
    }

    @Transactional
    override fun save(dto: GroupDto): GroupDto {
        throwIfNameAlreadyExists(dto.name)

        return super.save(dto).also {
            userLogic.saveDefaultGroupUser(it)
        }
    }

    override fun update(dto: GroupDto): GroupDto {
        throwIfNameAlreadyExists(dto.name, dto.id)

        return super.update(dto)
    }

    override fun deleteById(id: Long) {
        userLogic.deleteById(GroupDto.getDefaultGroupUserId(id))
        super.deleteById(id)
    }

    private fun throwIfNameAlreadyExists(name: String, id: Long? = null) {
        if (service.existsByName(name, id)) {
            throw alreadyExistsException(value = name)
        }
    }
}