package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.ServiceComponent
import dev.fyloz.colorrecipesexplorer.dtos.MaterialDto
import dev.fyloz.colorrecipesexplorer.logic.files.FileLogic
import dev.fyloz.colorrecipesexplorer.model.Material
import dev.fyloz.colorrecipesexplorer.repository.MaterialRepository
import org.springframework.beans.factory.annotation.Qualifier

interface MaterialService : Service<MaterialDto, Material, MaterialRepository> {
    /** Checks if a material with the given [name] and a different [id] exists. */
    fun existsByName(name: String, id: Long?): Boolean

    /** Updates the [inventoryQuantity] of the [Material] with the given [id]. */
    fun updateInventoryQuantityById(id: Long, inventoryQuantity: Float)

    /** Checks if a mix material or a mix type depends on the material with the given [id]. */
    fun isUsedByMixMaterialOrMixType(id: Long): Boolean
}

@ServiceComponent
class DefaultMaterialService(
    repository: MaterialRepository,
    private val materialTypeService: MaterialTypeService,
    @Qualifier("defaultFileLogic") val fileLogic: FileLogic
) :
    BaseService<MaterialDto, Material, MaterialRepository>(repository), MaterialService {
    override fun existsByName(name: String, id: Long?) = repository.existsByNameAndIdNot(name, id ?: 0)
    override fun getAll() = repository.findAllByIsMixTypeIsFalse().map(::toDto)
    override fun updateInventoryQuantityById(id: Long, inventoryQuantity: Float) =
        repository.updateInventoryQuantityById(id, inventoryQuantity)

    override fun isUsedByMixMaterialOrMixType(id: Long) = repository.isUsedByMixMaterialOrMixType(id)

    override fun toDto(entity: Material) =
        MaterialDto(
            entity.id,
            entity.name,
            entity.inventoryQuantity,
            entity.isMixType,
            materialTypeService.toDto(entity.materialType!!),
            hasSimdut(entity)
        )

    override fun toEntity(dto: MaterialDto) =
        Material(dto.id, dto.name, dto.inventoryQuantity, dto.isMixType, materialTypeService.toEntity(dto.materialType))

    private fun hasSimdut(material: Material) =
        fileLogic.exists("${Constants.FilePaths.SIMDUT}/${material.name}.pdf")
}