package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.dtos.EntityDto
import dev.fyloz.colorrecipesexplorer.model.ModelEntity
import org.springframework.data.jpa.repository.JpaRepository
import org.springframework.data.repository.findByIdOrNull

/**
 * Represents a service between the logic and the repository.
 * Gives access to the repository using a DTO.
 *
 * @param D The type of the entity DTO.
 * @param E The type of the entity.
 * @param R The repository of the entity.
 */
interface Service<D : EntityDto, E : ModelEntity, R : JpaRepository<E, Long>> {
    /** Checks if an entity with the given [id] exists. */
    fun existsById(id: Long): Boolean

    /** Gets all entities as DTOs. */
    fun getAll(): Collection<D>

    /** Gets the entity DTO with the given [id].*/
    fun getById(id: Long): D?

    /** Saves the given [dto]. */
    fun save(dto: D): D

    /** Deletes the given [dto]. */
    fun delete(dto: D)

    /** Deletes the entity with the given [id]. */
    fun deleteById(id: Long)

    /** Converts the given [entity] to a DTO. */
    fun toDto(entity: E): D

    /** Converts the given [dto] to an entity. */
    fun toEntity(dto: D): E
}

abstract class BaseService<D : EntityDto, E : ModelEntity, R : JpaRepository<E, Long>>(protected val repository: R) :
    Service<D, E, R> {
    override fun existsById(id: Long) =
        repository.existsById(id)

    override fun getAll() =
        repository.findAll().map(this::toDto)

    override fun getById(id: Long): D? {
        val entity = repository.findByIdOrNull(id) ?: return null
        return toDto(entity)
    }

    override fun save(dto: D): D {
        val entity = repository.save(toEntity(dto))
        return toDto(entity)
    }

    override fun delete(dto: D) {
        repository.delete(toEntity(dto))
    }

    override fun deleteById(id: Long) {
        repository.deleteById(id)
    }
}