package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.EntityDto
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.model.ModelEntity
import dev.fyloz.colorrecipesexplorer.service.Service
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.data.jpa.repository.JpaRepository
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

class BaseLogicTest {
    private val serviceMock = mockk<Service<TestEntityDto, ModelEntity, JpaRepository<ModelEntity, Long>>>()

    private val baseLogic = spyk(TestBaseLogic(serviceMock))

    private val dto = TestEntityDto(id = 1L)

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun existsById_normalBehavior_returnsTrue() {
        // Arrange
        every { serviceMock.existsById(any()) } returns true

        // Act
        val exists = baseLogic.existsById(dto.id)

        // Assert
        assertTrue(exists)
    }

    @Test
    fun exists_notFound_returnsFalse() {
        // Arrange
        every { serviceMock.existsById(any()) } returns false

        // Act
        val exists = baseLogic.existsById(dto.id)

        // Assert
        assertFalse(exists)
    }

    @Test
    fun getAll_normalBehavior_returnsAllDtos() {
        // Arrange
        val expectedDtos = listOf(dto)

        every { serviceMock.getAll() } returns expectedDtos

        // Act
        val actualDtos = baseLogic.getAll()

        // Assert
        assertEquals(expectedDtos, actualDtos)
    }

    @Test
    fun getById_normalBehavior_returnsDtoWithGivenId() {
        // Arrange
        every { serviceMock.getById(any()) } returns dto

        // Act
        val dtoById = baseLogic.getById(dto.id)

        // Assert
        assertEquals(dto, dtoById)
    }

    @Test
    fun getById_notFound_throwsNotFoundException() {
        // Arrange
        every { serviceMock.getById(any()) } returns null

        // Act
        // Assert
        assertThrows<NotFoundException> { baseLogic.getById(dto.id) }
    }

    @Test
    fun save_normalBehavior_callsServiceSave() {
        // Arrange
        every { serviceMock.save(any()) } returns dto

        // Act
        baseLogic.save(dto)

        // Assert
        verify {
            serviceMock.save(dto)
        }
        confirmVerified(serviceMock)
    }

    @Test
    fun save_normalBehavior_returnsSavedDto() {
        // Arrange
        every { serviceMock.save(any()) } returns dto

        // Act
        val savedDto = baseLogic.save(dto)

        // Assert
        assertEquals(dto, savedDto)
    }

    @Test
    fun update_normalBehavior_callsServiceSave() {
        // Arrange
        every { serviceMock.save(any()) } returns dto
        every { baseLogic.existsById(any()) } returns true

        // Act
        baseLogic.update(dto)

        // Assert
        verify {
            serviceMock.save(dto)
        }
        confirmVerified(serviceMock)
    }

    @Test
    fun update_normalBehavior_returnsUpdatedDto() {
        // Arrange
        every { serviceMock.save(any()) } returns dto
        every { baseLogic.existsById(any()) } returns true

        // Act
        val updatedDto = baseLogic.update(dto)

        // Assert
        assertEquals(dto, updatedDto)
    }

    @Test
    fun update_notFound_throwsNotFoundException() {
        // Arrange
        every { serviceMock.save(any()) } returns dto
        every { baseLogic.existsById(any()) } returns false

        // Act
        // Assert
        assertThrows<NotFoundException> { baseLogic.update(dto) }
    }

    @Test
    fun deleteById_normalBehavior_callsServiceDeleteById() {
        // Arrange
        every { serviceMock.deleteById(any()) } just runs

        // Act
        baseLogic.deleteById(dto.id)

        // Assert
        verify {
            serviceMock.deleteById(dto.id)
        }
        confirmVerified(serviceMock)
    }
}

private data class TestEntityDto(override val id: Long) : EntityDto
private class TestBaseLogic<D : EntityDto, S : Service<D, *, *>>(service: S) :
    BaseLogic<D, S>(service, "UnitTestType")