package dev.fyloz.colorrecipesexplorer.logic.account

import dev.fyloz.colorrecipesexplorer.dtos.GroupDto
import dev.fyloz.colorrecipesexplorer.dtos.UserDto
import dev.fyloz.colorrecipesexplorer.dtos.UserSaveDto
import dev.fyloz.colorrecipesexplorer.dtos.UserUpdateDto
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.logic.users.DefaultUserLogic
import dev.fyloz.colorrecipesexplorer.logic.users.GroupLogic
import dev.fyloz.colorrecipesexplorer.model.account.Permission
import dev.fyloz.colorrecipesexplorer.service.UserService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.security.crypto.password.PasswordEncoder
import java.time.LocalDateTime

class DefaultUserLogicTest {
    private val user = UserDto(1L, "Unit test", "User", "asecurepassword", null, listOf())
    private val group = GroupDto(1L, "Unit test group", listOf())

    private val userServiceMock = mockk<UserService> {
        every { existsById(any()) } returns false
        every { existsByFirstNameAndLastName(any(), any(), any()) } returns false
        every { getAll(any(), any()) } returns listOf()
        every { getAllByGroup(any()) } returns listOf()
        every { getById(any(), any(), any()) } returns user
        every { getByFirstNameAndLastName(any(), any()) } returns user
        every { getDefaultGroupUser(any()) } returns user
    }
    private val groupLogicMock = mockk<GroupLogic> {
        every { getById(any()) } returns group
    }
    private val passwordEncoderMock = mockk<PasswordEncoder> {
        every { encode(any()) } answers { "encoded ${this.firstArg<String>()}" }
    }

    private val userLogic = spyk(DefaultUserLogic(userServiceMock, groupLogicMock, passwordEncoderMock))

    private val userSaveDto = UserSaveDto(
        user.id,
        user.firstName,
        user.lastName,
        user.password,
        null,
        user.permissions,
        user.isSystemUser,
        user.isDefaultGroupUser
    )
    private val userUpdateDto = UserUpdateDto(user.id, user.firstName, user.lastName, null, listOf())

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    @Test
    fun getAll_normalBehavior_callsGetAllInServiceWithSpecialUsersDisabled() {
        // Arrange
        // Act
        userLogic.getAll()

        // Assert
        verify {
            userServiceMock.getAll(isSystemUser = false, isDefaultGroupUser = false)
        }
        confirmVerified(userServiceMock)
    }

    @Test
    fun getAllByGroup_normalBehavior_callsGetAllByGroupInService() {
        // Arrange
        // Act
        userLogic.getAllByGroup(group)

        // Assert
        verify {
            userServiceMock.getAllByGroup(group)
        }
        confirmVerified(userServiceMock)
    }

    @Test
    fun getById_default_normalBehavior_callsGetByIdWithSpecialUsersDisabled() {
        // Arrange
        // Act
        userLogic.getById(user.id)

        // Assert
        verify {
            userLogic.getById(user.id, isSystemUser = false, isDefaultGroupUser = false)
        }
    }

    @Test
    fun getById_normalBehavior_callsGetByIdInService() {
        // Arrange
        // Act
        userLogic.getById(user.id, isSystemUser = false, isDefaultGroupUser = true)

        // Assert
        verify {
            userServiceMock.getById(user.id, isSystemUser = false, isDefaultGroupUser = true)
        }
        confirmVerified(userServiceMock)
    }

    @Test
    fun getById_notFound_throwsNotFoundException() {
        // Arrange
        every { userServiceMock.getById(any(), any(), any()) } returns null

        // Act
        // Assert
        assertThrows<NotFoundException> { userLogic.getById(user.id) }
    }

    @Test
    fun getDefaultGroupUser_normalBehavior_callsGetDefaultGroupUserInService() {
        // Arrange
        // Act
        userLogic.getDefaultGroupUser(group)

        // Assert
        verify {
            userServiceMock.getDefaultGroupUser(group)
        }
        confirmVerified(userServiceMock)
    }

    @Test
    fun getDefaultGroupUser_notFound_throwsNotFoundException() {
        // Arrange
        every { userServiceMock.getDefaultGroupUser(any()) } returns null

        // Act
        // Assert
        assertThrows<NotFoundException> { userLogic.getDefaultGroupUser(group) }
    }

    @Test
    fun saveDefaultGroupUser_normalBehavior_callsSaveWithValidSaveDto() {
        // Arrange
        every { userLogic.save(any<UserSaveDto>()) } returns user

        val expectedSaveDto = UserSaveDto(
            group.defaultGroupUserId, group.name, "User", group.name, group.id, listOf(), isDefaultGroupUser = true
        )

        // Act
        userLogic.saveDefaultGroupUser(group)

        // Assert
        verify {
            userLogic.save(expectedSaveDto)
        }
    }

    @Test
    fun save_dto_normalBehavior_callsSaveWithValidUser() {
        // Arrange
        every { userLogic.save(any<UserDto>()) } returns user

        val expectedUser = user.copy(password = "encoded ${user.password}")

        // Act
        userLogic.save(userSaveDto)

        // Assert
        verify {
            userLogic.save(expectedUser)
        }
    }

//    TODO Causes a stackoverflow because of a bug in mockk
//    @Test
//    fun save_normalBehavior_callsSaveInService() {
//        // Arrange
//        // Act
//        userLogic.save(user)
//
//        // Assert
//        verify {
//            userServiceMock.save(user)
//        }
//    }

    @Test
    fun save_idAlreadyExists_throwsAlreadyExistsException() {
        // Arrange
        every { userServiceMock.existsById(any()) } returns true

        // Act
        // Assert
        assertThrows<AlreadyExistsException> { userLogic.save(user) }
    }

    @Test
    fun save_fullNameAlreadyExists_throwsAlreadyExistsException() {
        // Arrange
        every { userServiceMock.existsByFirstNameAndLastName(any(), any(), any()) } returns true

        // Act
        // Assert
        assertThrows<AlreadyExistsException> { userLogic.save(userSaveDto) }
    }

    @Test
    fun update_dto_normalBehavior_callsUpdateWithValidUser() {
        // Arrange
        every { userLogic.getById(any(), any(), any()) } returns user
        every { userLogic.update(any<UserDto>()) } returns user

        // Act
        userLogic.update(userUpdateDto)

        // Assert
        verify {
            userLogic.update(user)
        }
    }

    @Test
    fun update_fullNameAlreadyExists_ThrowAlreadyExistsException() {
        // Arrange
        every { userServiceMock.existsByFirstNameAndLastName(any(), any(), any()) } returns true

        // Act
        // Assert
        assertThrows<AlreadyExistsException> { userLogic.update(user) }
    }

    @Test
    fun updateLastLoginTime_normalBehavior_callsUpdateWithUpdatedTime() {
        // Arrange
        every { userLogic.getById(any()) } returns user
        every { userLogic.update(any<UserDto>()) } returns user

        val time = LocalDateTime.now()
        val expectedUser = user.copy(lastLoginTime = time)

        // Act
        userLogic.updateLastLoginTime(user.id, time)

        // Assert
        verify {
            userLogic.update(expectedUser)
        }
    }

    @Test
    fun updatePassword_normalBehavior_callsUpdateWithUpdatedTime() {
        // Arrange
        every { userLogic.getById(any()) } returns user
        every { userLogic.update(any<UserDto>()) } returns user

        val updatedPassword = "updatedpassword"
        val expectedUser = user.copy(password = "encoded $updatedPassword")

        // Act
        userLogic.updatePassword(user.id, updatedPassword)

        // Assert
        verify {
            userLogic.update(expectedUser)
        }
    }

    @Test
    fun addPermission_normalBehavior_callsUpdateWithAddedPermission() {
        // Arrange
        every { userLogic.getById(any()) } returns user
        every { userLogic.update(any<UserDto>()) } returns user

        val addedPermission = Permission.VIEW_COMPANY
        val expectedUser = user.copy(permissions = user.permissions + addedPermission)

        // Act
        userLogic.addPermission(user.id, addedPermission)

        // Assert
        verify {
            userLogic.update(expectedUser)
        }
    }

    @Test
    fun removePermission_normalBehavior_callsUpdateWithAddedPermission() {
        // Arrange
        val removedPermission = Permission.VIEW_COMPANY
        val baseUser = user.copy(permissions = user.permissions + removedPermission)

        every { userLogic.getById(any()) } returns baseUser
        every { userLogic.update(any<UserDto>()) } returns user

        // Act
        userLogic.removePermission(user.id, removedPermission)

        // Assert
        verify {
            userLogic.update(user)
        }
    }
}